

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
//
// Features of Rock's Server Admin:
//
//	Administration is performed from the admin-player's remote
//	game-playing machine.  The admin-player need NOT be located
//	at the server to perform the admin duties.
//
//
//	Automatic password control of admin capabilities.  The admin-player
//	takes control from his remote game-playing machine.  Control is
//	automatically disabled if the admin-player leaves the game.
//
//
//	Automatic routing of admin information to ONLY the admin-player's
//	chat window.
//
//
//	Automatic Banned-IP list verification:  Checks new players against
//	a file of permanently banned IPs and automatically gives him
//	Rock's Big Boot if there's a match.
//
//	
//	Automatic addition of a player to the Banned-IP list.  You do not
//	have to edit files or scripts to add a player to your Ban List.
//
//	
//	Automatic logging of game activity with either Rock's Game Log,
//	or Starsiege's built-in game log facility.
//
//
//	Automatic display of player's IP address in the admin-player's chat
//	window as each player enters the game.
//
//
//	Automatic detection and display of a stealth-player when they enter
//	or exit a game.
//
//
//	Automatic detection and monitoring for R99 server crashers.  Upon
//	detection of an offending player, the package automatically adds
//	the player's IP adr to the Banned-IP list and bans the player from
//	the server.
//
//
//	Automatic detection of unknown players. Politely rejects player-entry
//	to those with unknown as a name.
//
//
//	Automatic detection and monitoring for appropriate color-tags.  Upon
//	detection of an improper color-tag, the package ejects the player
//	with a polite messageBox indicating his need to change his tags.
//
//
//	Automatic detection and monitoring for Wait Room camping idle players
//	to prevent unnecessay connection bandwidth consumption.  When detected,
//	the player is politely ejected with an indicating messageBox.
//
//
//	Automatic Team assignment of players when in TDM mode.  Permits squads
//	to automatically stay on one team.
//
//
//	Admin actions can be applied from the Wait Room because player
//	selection is NOT done via targeting.
//
//
//	Auto-Heal after a fight, immediate or delayed.
//
//
//	Graceful Server Resetting at the end of the Mission List.
//
//
//	Admin-actions applied via Remote Control from the admin's machine:
//
//		Issuing an automated warning to an unruly player
//
//		Kicking naughty players
//
//		Banning badboy players
//
//		Ping a player from in-game
//
//		Cycling of the game to the next map
//
//		Music Off or play custom music groups.
//
//		Reassign admin capabilities to another player
//
//		Reaquire admin capabilites from a player.
//
//		Enable/Disable the Auto-Heal Algorithm
//
//		Enable/Disable the Game Logging facility
//
//		Enable/Disable ATR-style of scoring
//
//		Enable/Disable of the R99 Monitor
//
//		Enable/Disable of the Tag Monitor
//
//		Enable/Disable of the NoUnk detector
//
//		Display player info for selected player
//
//		Display player info for all players
//
//		Add a player to the Ban List
//
//		Reconfig and restart the Game Server
//
//		Enable/Disable AutoTeam
//
//
//
//	Modular design, well documented scripts, and easy retrofit to
//	your server.  The scripts don't contribute any noticable lag.
//
//
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------




// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
//
// General Notes:
// The functions in this file have to load once with every initial
// server-start.  Therefore, this file should be executed from within
// the initial server file such as Server_DM_ALL.cs or its equivalent.
//
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------




// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// Globals used for general housekeeping.  DON'T MESS WITH THESE!
//
$RocksGameCycle			= FALSE;

$RocksRestartFlag		= FALSE;

$RocksServerCycle		= 0;

$RocksMusicSwitch		= 0;

$RocksMissionFlag		= 0;

$RocksMapNumber			= 0;

$RocksServerFlipperFile		= $RocksInstallDir @ "\\RocksServerFlipper.cs";

$RocksAutoConfigFile		= "RocksAutoConfig.cs";

$RocksAutoTeamMode		= 0;

$RocksVersion			= "4.2";




// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// ------------------------- Admin Functions --------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// Enable/Disable all vehicles and weapons.
//
function RocksAllowAll(%onoff)
{
	allowVehicle(all,   false);	// assure all hack vehicles are off
	allowComponent(all, true);
	allowWeapon(all,    %onoff);

	RocksHeavies(%onoff);
	RocksLights(%onoff);
	RocksTanks(%onoff);
	RocksDisrupters(%onoff);
}


// --------------------------------------------------------------------------
// Enable/Disable all 60-ton and over Hercs, except tanks
//
function RocksHeavies(%onoff)
{
	allowVehicle(   1,  %onoff);  // Terran Apocalypse                
	allowVehicle(   3,  %onoff);  // Terran Gorgon                    
	allowVehicle(  10,  %onoff);  // Knight's Apocalypse
	allowVehicle(  12,  %onoff);  // Knight's Gorgon  
	allowVehicle(  33,  %onoff);  // Rebel Olympian 
	allowVehicle(  23,  %onoff);  // Cybrid Adjudicator  
	allowVehicle(  24,  %onoff);  // Cybrid Executioner  
	allowVehicle(  55,  %onoff);  // Platinum Adjudicator 2
	allowVehicle(  56,  %onoff);  // Platinum Executioner 2
}


// --------------------------------------------------------------------------
// Enable/Disable all Hercs less than 60-tons, except tanks
//
function RocksLights(%onoff)
{
	allowVehicle(  20,  %onoff);  // Seeker
	allowVehicle(  35,  %onoff);  // Metagen Seeker
	allowVehicle(  21,  %onoff);  // Goad
	allowVehicle(  36,  %onoff);  // Metagen Goad
	allowVehicle(  22,  %onoff);  // Shepherd
	allowVehicle(  37,  %onoff);  // Metagen Shepherd
	allowVehicle(   4,  %onoff);  // Talon
	allowVehicle(  13,  %onoff);  // Knight's Talon
	allowVehicle(   2,  %onoff);  // Minotaur
	allowVehicle(  11,  %onoff);  // Knight's Minotaur
	allowVehicle(   5,  %onoff);  // Basilisk
	allowVehicle(  14,  %onoff);  // Knight's Basilisk
	allowVehicle(  30,  %onoff);  // Emancipator
	allowVehicle(  52,  %onoff);  // Outrider Emancipator
}


// --------------------------------------------------------------------------
// Enable/Disable all missiles, and other guided weapons
//
function RocksMissiles(%onoff)
{
	allowWeapon(  110,  %onoff);	// Plasma
	allowWeapon(  124,  %onoff);	// Viper 8
	allowWeapon(  125,  %onoff);	// Viper 12
	allowWeapon(  126,  %onoff);	// Sparrow 6
	allowWeapon(  127,  %onoff);	// Sparrow 10
	allowWeapon(  128,  %onoff);	// Swarm 6
	allowWeapon(  129,  %onoff);	// Minion
	allowWeapon(  130,  %onoff);	// Shrike
	allowWeapon(  131,  %onoff);	// Arachnitron 4
	allowWeapon(  132,  %onoff);	// Arachnitron 8
	allowWeapon(  133,  %onoff);	// Arachnitron 12
	allowWeapon(  147,  %onoff);	// Aphid
	allowWeapon(  150,  %onoff);	// SGuns
}


// --------------------------------------------------------------------------
// Enable/Disable all mines
//
function RocksMines(%onoff)
{
	allowWeapon(  134,  %onoff);	// Mine 6
	allowWeapon(  135,  %onoff);	// Mine 10
	allowWeapon(  136,  %onoff);	// Mine 15
}


// --------------------------------------------------------------------------
// Enable/Disable all tanks except Disrupters.
//
function RocksTanks(%onoff)
{
	allowVehicle(   6,  %onoff);  // Paladin Tank                     
	allowVehicle(   7,  %onoff);  // Myrmidon Tank                    
	allowVehicle(  15,  %onoff);  // Knight's Paladin 
	allowVehicle(  16,  %onoff);  // Knight's Myrmidon
	allowVehicle(  25,  %onoff);  // Bolo Tank           
	allowVehicle(  26,  %onoff);  // Recluse Tank
	allowVehicle(  31,  %onoff);  // Avenger Tank   
	allowVehicle(  32,  %onoff);  // Dreadnought Tank
	allowVehicle(  41,  %onoff);  // Harabec's Predator
}


// --------------------------------------------------------------------------
// Enable/Disable just Disrupters.
//
function RocksDisrupters(%onoff)
{
	allowVehicle(   8,  %onoff);  // Disruptor Tank
	allowVehicle(  17,  %onoff);  // Knight's Disruptor
}


//----------------------------------------------------------------------------
// This makes admin-announcements to all of the other Admin people.  This is
// used because team-messages are sent to vehicles, not players, and Admin
// players need not be in a vehicle on the field.  Most Admin messages should
// be announced to all Admin to help keep them informed of activity.
//
function RocksAnnounce(%str)
{
	%count = playerManager::getPlayerCount();

	for (%i = 0; %i < %count; %i++) {
		%player = playerManager::getPlayerNum(%i);

		if (%player && %player.admin)
			say(%player, 0, %str);
	}
}


// --------------------------------------------------------------------------
// This makes a player information string so that all info will have the
// same standardized format
//
function RocksInfoMaker(%player)
{
	// It turns out that Starsiege's getPing() function always
	// returns 0 so I don't bother with the ping info since it
	// is obviously turned off or has a bug in it.
	//
	//	return("\t" @ %player.ipAdr @ " P:" @
	//		getPing(%player) @ " \t" @ getName(%player));
	//

	
	%info = getName(%player);


	// This bumps the field out so they all line up
	//
	for (%i = strlen(%info); %i < 24; %i += 8)
		%info = %info @ "\t";


	return("\t" @ %info @ %player.ipAdr);
}


// --------------------------------------------------------------------------
// Admin players gain access here.  This implements the NUTS idea of
// permitting more than one Admin at a time, each having Admin control.
//
function RocksControlGrabber(%player)
{
	%who = getName(%player);

	%info = RocksInfoMaker(%player);

	%vehicle = playerManager::playerNumToVehicleId(%player);


	if (%player.admin) {

		RocksAnnounce("\t" @ %who @ " has yielded Admin Control.");

		if (!%vehicle)
			%player.idle = $RocksIdleTime;
		else if ($server::TeamPlay == false)
			setTeam(%vehicle, *IDSTR_TEAM_RED);

		%player.admin = 0;

		RocksLogEntry("\tAdmin-" @ %info);

		return;
	}


	%player.idle = 0;
	%player.admin = 1;

	if (%vehicle && ($server::TeamPlay == false))
		setTeam(%vehicle, $RocksAdminColor);


	RocksAnnounce("\t" @ %who @ " has acquired Admin Control." @
				"  V" @ $RocksVersion);

	RocksLogEntry("\tAdmin+" @ %info);
}


// --------------------------------------------------------------------------
// Displays one or all player's IP addresses in the Admin's chat window.
//
function RocksShowPlayerInfo(%admin, %player, %one)
{
	if (%one) {
		%info = RocksInfoMaker(%player);

		if (!%admin)
			RocksAnnounce(%info);
		else {
			if (%player.admin)
				%info = %info @ "  - Admin";

			if (%player == %admin.selectPlayer)
				%info = %info @ "  (Selected)";

			say(%admin, 0, %info);
		}

		return;
	}


	// Note that this loop is recursive
	//
	%count = playerManager::getPlayerCount();

	for (%i = 0; %i < %count; %i++) {
		%player = playerManager::getPlayerNum(%i);

		if (%player)
			RocksShowPlayerInfo(%admin, %player, 1);
	}
}


// --------------------------------------------------------------------------
// Implements a simple mathematical Random Number Generator (RNG) which has
// better random dispersion than the game's built in randonInt() routine.
// The core-function returns a random integer in the range 1 to limit. Once
// the seed if first initialized, I overload the game's built-in randomInt()
// function to use this one instead.
//
// Although not necessary, you can reseed the RNG with a new value with
// a single call to RocksRNGseed().  Because the call uses the getSimTime()
// function, you should wait until after the server is running before making
// the call.  Again, it is not necessary to continuously reseed the RNG.
//
// To prevent math-overflow, I cap the quantity (max - min) to be less than
// or equal to 0x7FFF (32767).
//
$RNGSeed = randomInt(1,10000);	// init the seed with the game's function

function RocksRNG(%limit)
{
	%limit %= 0x7FFF;

	$RNGSeed = (($RNGSeed * 32719) + 3) % 32749;

	return(($RNGSeed % %limit) + 1);
}

function RocksRNGseed()
{
	%value = floor(getSimTime() * 1001);

	if (%value > 0)
		$RNGSeed = RocksRNG(%value);
}

// Overload the game's built-in RNG so we use RocksRNG() instead.
//
function randomInt(%min, %max)
{
	// normalize the range to 1
	//
	return(RocksRNG((%max - %min) + 1) + (%min - 1));
}


// --------------------------------------------------------------------------
// Generic player discipline routine.  This is used to enforce the rules
// of the server on a dynamic basis.  The messages are fairly benign when
// compared to the one in RocksBigFoot().  The level indicates the intensity
// of the discipline.
//
function RocksFoot(%admin, %player, %level)
{
	%name = " (" @ getName(%admin) @ ")";

	%info = RocksInfoMaker(%player) @ %name;

	if (%level == 1) {			// Issue warning

		%logstr = "\tWarned" @ %info;

		%str2 = "\n" @ getName(%player) @
			"\n\nPlease read and abide by the rules\nlisted " @
			"in the Wait Room Rules Tab.\n- Administrator -";

		messageBox (%player, %str2);

		%vehicle = playerManager::playerNumTovehicleId(%player);

		if (%vehicle) {
			violate(%vehicle);
			say (%player, 0, "\t" @ %str2, "Rules_violated.wav");
		}
		else say (%player, 0, "\t" @ %str2);
	}


	else if (%level == 2) {			// Kick the unruly player

		%logstr = "\tKicked" @ %info;

		kick(%player, "\n\nAdministrative Kick");
	}


	else if (%level == 3) {			// Ban that Bad Boy

		%logstr = "\tBanned" @ %info;

		ban(%player, 30, "\n\nAdministrative Ban");
	}


	RocksLogEntry(%logstr);

	RocksAnnounce(%logstr);
}


// --------------------------------------------------------------------------
// This is a rotational selector that selects one player from the game without
// the Admin having to be on the field to target a player.  As a convenience,
// the routine rotates forward or backwards thru the player list.
//
function RocksPlayerSelect(%admin, %forward)
{
	%count = playerManager::getPlayerCount();


	// Performs forward or backward rotation thru the player list.
	// Normally I would do this with mod-arithmetic, but since everything
	// is in floating point, it's a bad idea.
	//
	if (%forward) {
		%admin.selectIndex++;
		if (%admin.selectIndex >= %count)
			%admin.selectIndex = 0;
	}
	else {
		%admin.selectIndex--;
		if (%admin.selectIndex < 0)
			%admin.selectIndex = %count-1;
	}

	%admin.selectPlayer = playerManager::getPlayerNum(%admin.selectIndex);
	%admin.selectVehicle = 0;

	if (%admin.selectPlayer)
		RocksShowPlayerInfo(%admin, %admin.selectPlayer, 1);
}


// --------------------------------------------------------------------------
// This algorithm sets one or more fields of an IP address to zero and returns
// the edited address.  This is because when we compare an IP to another,
// we have to have some fields, such as the port number, set to zero.  This
// is a somewhat tricky algo because Starsiege's strAlign() function is only
// useful from the left, not the right.  This means that to find a character
// in the IP address, we have to perform a look-ahead algorithm.
//
// Note that an IP adr has 5 fields.  For proper operation of the Ban List
// comparison function, we have to set lesser subfields to zero so that the
// isEqualIP() routine will ignore some subfields.  The %field parameter tells
// how many fields to return.
//
//	Fields	    1   2  3  4    5
//		IP:192.168.1.100:12345
//
// Any number of fields other than 1 thru 4 will return the unedited IP
//
function RocksGetAddressIP(%player, %field)
{
	%adr = getConnection(%player);

	if ((%field < 1) || (%field >= 5) || (%adr == "LOOPBACK"))
		return(%adr);

	%count = strlen(%adr);

	%fieldcount = 1;

	%last = strAlign(3, left, %adr);

	// start at 4 because the first field is at least that long
	//
	for (%i = 4; %i < %count; %i++) {

		%ip = strAlign(%i, left, %adr);

		if (%ip == (%last @ ".")) {

			if (%fieldcount == %field) {
				for (%i = %fieldcount; %i < 4; %i++) {
					if (%i < 3)
						%ip = %ip @ "0.";
					else %ip = %ip @ "0";
				}

				return(%ip @ ":0");
			}

			%fieldcount++;
		}
		else if (%ip == (%last @ ":"))
			return(%ip @ "0");

		%last = %ip;
	}
}


// --------------------------------------------------------------------------
// Inits the ASCII character table with hex representations, and also sets up
// upper-case and lower-case optimization arrays.
//
$hex[ 0] = 0; $hex[ 1] = 1; $hex[ 2] = 2; $hex[ 3] = 3;
$hex[ 4] = 4; $hex[ 5] = 5; $hex[ 6] = 6; $hex[ 7] = 7;
$hex[ 8] = 8; $hex[ 9] = 9; $hex[10] = a; $hex[11] = b;
$hex[12] = c; $hex[13] = d; $hex[14] = e; $hex[15] = f;

function RocksAsciiInit()
{
	for (%i = 0; %i < 256; %i++) {
		%hi = %i / 16;
		if (%hi >= 10)
			%hi = strAlign(2, left, %hi);
		else %hi = strAlign(1, left, %hi);

		%lo = %i % 16;

		eval("$ascii[%i] = \"\\x" @ $hex[%hi] @ $hex[%lo] @ "\";");
	}

	for (%i = 0; %i < 9; %i++) {
		eval("$upper[%i] = \"\\x" @ 41+%i @ "\";");
		eval("$lower[%i] = \"\\x" @ 61+%i @ "\";");
	}

	$upper[ 9] = "\x4a"; $lower[ 9] = "\x6a";
	$upper[10] = "\x4b"; $lower[10] = "\x6b";
	$upper[11] = "\x4c"; $lower[11] = "\x6c";
	$upper[12] = "\x4d"; $lower[12] = "\x6d";
	$upper[13] = "\x4e"; $lower[13] = "\x6e";
	$upper[14] = "\x4f"; $lower[14] = "\x6f";

	for (%i = 15; %i < 25; %i++) {
		eval("$upper[%i] = \"\\x" @ 50 + (%i - 15) @ "\";");
		eval("$lower[%i] = \"\\x" @ 70 + (%i - 15) @ "\";");
	}

	$upper[25] = "\x5a"; $lower[25] = "\x7a";
}


// --------------------------------------------------------------------------
// Converts str to all lowercase characters and returns the new string.
//
function RocksStrToLower(%str)
{
	%new = "";

	%str = convertSpaces(%str);
	%count = strlen(%str);

	for (%i = 0; %i < %count; %i++) {

		%sample1 = strAlign(%i, left, %str);
		%sample2 = strAlign(%i+1, left, %str);

		// find the character
		//
		for (%j = 0; %j < 26; %j++) {
			if (((%sample1 @ $upper[%j]@"#") == %sample2@"#") ||
			    ((%sample1 @ $lower[%j]@"#") == %sample2@"#")) {
				%new = %new @ $lower[%j];
				break;
			}
		}

		if (%j != 26) continue;

		for (%j = 0; %j < 256; %j++) {
			if ((%sample1 @ $ascii[%j]@"#") == %sample2@"#") {
				%new = %new @ $ascii[%j];
				break;
			}
		}
	}

	return(%new);
}


// --------------------------------------------------------------------------
// This performs a look-ahead algorithm to search a string for a substr.  We
// have to tack a #-character on the end because a '.' 'e' '-' and '+' will
// all match almost anything... I don't know why except that they are math
// modifiers.  Therefore, I create an artificial character with the #-char.
// Returns the total number of substrs within str.
//
function RocksSubStr(%str, %substr)
{
	%total = 0;

	%count1 = strlen(%str);

	%count2 = strlen(%substr);

	if ((%count1 < %count2) || (!%count1) || (!%count2))
		return(0);

	%count1 -= (%count2 - 1);

	%substr = %substr @ "#";

	for (%i = 0; %i < %count1; %i++) {

		%new = strAlign(%i, left, %str);

		if (%new @ %substr == (strAlign(%i+%count2, left, %str)@"#"))
			%total++;
	}

	return(%total);
}


// --------------------------------------------------------------------------
// Given a string, this returns the right-most N characters.  If N is greater
// than or equal to the string-length, then the string is returned unmodified,
// except that spaces will be converted to underscores.
//
function RocksStrRight(%str, %n)
{
	%str = convertSpaces(%str);

	%count = strlen(%str);

	if (%n >= %count)
		return(%str);

	if (%n <= 0)
		return("");

	%count -= %n;


	// get everything we're going to lop off
	//
	%sub = strAlign(%count, left, %str);

	%final = "";

	for (%i = 0; %i < %n; %i++) {

		%count++;

		%sample = strAlign(%count, left, %str) @ "#";

		for (%j = 0; %j < 256; %j++) {

			if ((%sub @ $ascii[%j] @ "#") == %sample) {

				%sub = %sub @ $ascii[%j];

				%final = %final @ $ascii[%j];

				break;
			}
		}
	}

	return(%final);
}


// --------------------------------------------------------------------------
// You can thank GotLag? for how this pair of functions work.
//
function RocksPingPlayer(%admin, %player)
{
	%player.pingStart = getSimTime();

	remoteEval(%player, "eval", "PingReturn", %admin);
}

function remotePingReturn(%player, %admin)
{
	%player.pingEnd = getSimTime();

	say (%admin, 0, "\tPing " @
		floor(((%player.pingEnd - %player.pingStart) * 1000) + 0.5) @ 
		": " @ RocksInfoMaker(%player));
}


// --------------------------------------------------------------------------
// Finds out which player another player has targeted.
//
function RocksTargetedPlayer(%player)
{
	%vehicle = playerManager::playerNumTovehicleId(%player);

	if (!%vehicle)
		return(0);

	%target = getCurrentTargetID(%vehicle);

	if (!%target)
		return(0);

	return(playerManager::vehicleIdToPlayerNum(%target));
}


// --------------------------------------------------------------------------
// Finds out which vehicle another player has targeted.
//
function RocksTargetedVehicle(%player)
{
	%vehicle = playerManager::playerNumTovehicleId(%player);

	if (!%vehicle)
		return(0);

	return(getCurrentTargetID(%vehicle));
}


// --------------------------------------------------------------------------
// Switches the CD music group every iteration, or on command
//
function RocksCDAudio(%player, %cmd)
{
	if (%cmd == 0) {
		if ($RocksMusicSwitch == 0)
			$RocksMusicSwitch = 1;

		if ($RocksMusicSwitch > 0) {
			$RocksMusicSwitch = -$RocksMusicSwitch;
			cdAudioFadeStop();
		}

		say (%player, 0, "\tMusic Off");
		return;
	}


	if ((%cmd == 1) && ($RocksMusicSwitch < 0))
		$RocksMusicSwitch = -$RocksMusicSwitch;
	else {
		$RocksMusicSwitch++;
		if ($RocksMusicSwitch >= 4)
			$RocksMusicSwitch = 1;

		cdAudioFadeStop();
	}


	if ($RocksMusicSwitch == 1) {
		say (%player, 0, "\tWatching Gnash Newtech Yougot");
		cdAudioCycle("Watching", "Gnash", "Newtech", "Yougot");
	}
	else if ($RocksMusicSwitch == 2) {
		say (%player, 0, "\tCloudBurst Cyberntx Mechsoul Terror");
		cdAudioCycle("CloudBurst", "Cyberntx", "Mechsoul", "Terror");
	}
	else if ($RocksMusicSwitch == 3) {
		say (%player, 0, "\tSS4 SS3 Purge");
		cdAudioCycle("SS4", "SS3", "Purge");
	}
}


// --------------------------------------------------------------------------
// Determines the Elapsed Time between starttime and endtime and returns a
// string in the form of hrs:min:sec with any necessary leading zeros to fill
// each field out to 2 digits.  Since arithmetic in the game is always done
// in floating point, we have to do this as integer-math.
//
function RocksET(%starttime, %endtime)
{
	%diff = %endtime - %starttime;

	%hours = 0;
	%minutes = 0;

	%temp = %diff - 3600;

	while (%temp >= 0) {
		%hours++;
		%diff -= 3600;
		%temp -= 3600;
	}

	if (%hours < 10)
		%hours = "0" @ %hours;


	%temp = %diff - 60;

	while (%temp >= 0) {
		%minutes++;
		%diff -= 60;
		%temp -= 60;
	}

	if (%minutes < 10)
		%minutes = "0" @ %minutes;


	if (%diff < 10)
		%diff = "0" @ %diff;

	return(%hours @ ":" @ %minutes @ ":" @ %diff);
}


// --------------------------------------------------------------------------
// Implements the ATR-style of scoring.  This scoring applies points
// according to the following formula: (kills / (deaths + 1)) * 100
// Therefore, the scoring stresses a high kills/deaths ratio, thus rewarding
// survival and combat abilities to stay alive, instead of the number of kills.
//
// This also fixes the team scoring where the team-score doesn't change when
// a player leaves the server.
//
function RocksPlayerScore(%a)
{
	%score =  EXTPlayerScore(%a);
	if (%score != 9999)
		return(%score);


	if ($RocksATRscoring)
		return (getKills(%a) / (getDeaths(%a) + 1) * 100);
	else return ((getKills(%a) * $killPoints) -
			(getDeaths(%a) * $deathPoints));
}

function RocksTeamScore(%a)
{
	%score =  EXTTeamScore(%a);
	if (%score != 9999)
		return(%score);


	%name = getTeamNameFromTeamID(%a);

	if ($RocksATRscoring) {
		return ($RocksTDMKills[%name] /
			($RocksTDMDeaths[%name] + 1) * 100);
	}
	else {
		return (($RocksTDMKills[%name] * $killPoints) -
			($RocksTDMDeaths[%name] * $deathPoints));
	}
}

function RocksTeamKills(%a)
{
	return ($RocksTDMKills[getTeamNameFromTeamID(%a)]);
}

function RocksTeamDeaths(%a)
{
	return ($RocksTDMDeaths[getTeamNameFromTeamID(%a)]);
}


// --------------------------------------------------------------------------
// Initializes the password.  If there is anything out of order with the
// loading of the password, then a random number is assigned, effectively
// locking out all admin capabilities.  This function is duplicated in the
// client machine's keybindings file so that the same password file can be
// used on the client.
//
function RocksPasswordInit()
{
	$AdminPassword = "";

	if (isFile($RocksInstallDir @ "\\" @ $RocksPasswordFile))
		exec($RocksInstallDir @ "\\" @ $RocksPasswordFile);

	if ($AdminPassword == "")
		$AdminPassword = randomInt(1, 10000);
}


// --------------------------------------------------------------------------
// Since onMissionLoad() is overloaded by the mission-file after the Admin
// hooks are placed, I can't hook it in the RocksDMLib.cs file.  So I do
// this round-about way of detecting a new mission.
//
function RocksMissionStart(%player)
{

	if (($CmdLineServer == TRUE) && ($RocksMapNumber == $RocksMapCount)) {

		if (%player || ($RocksRestartFlag == FALSE)) {
			say(%player, 0,
				"\tWARNING: Server will Auto-Restart " @
				"at end of this mission");
		}

		if (!%player && ($RocksRestartFlag == FALSE)) {

			$RocksRestartFlag = TRUE;

			// Schedule the restart warnings
			//
			%limit = $server::TimeLimit * 60;

			RocksRestartWarning(%limit);

			// This assures that the Restart Key-Command doesn't
			// conflict with the end-of-maplist restart warnings.
			//
			schedule("$RocksServerCycle++;", %limit - 60);
		}
	}

	if ($RocksMissionFlag)
		return;

	$RocksMissionFlag = 1;


	// Start the Handle Monitor if we're configured
	// for it and it's not already running.
	//
	if ($RocksR99Enable || $RocksTagEnable || ($RocksIdleTime > 0))
		RocksMonitor(1);


	// Start Random Thoughts
	//
	if ($RocksRandomThoughts)
		RRTStart();


	$RocksTDMKills[getTeamNameFromTeamID(1)] = 0;
	$RocksTDMKills[getTeamNameFromTeamID(2)] = 0;
	$RocksTDMKills[getTeamNameFromTeamID(4)] = 0;
	$RocksTDMKills[getTeamNameFromTeamID(8)] = 0;

	$RocksTDMDeaths[getTeamNameFromTeamID(1)] = 0;
	$RocksTDMDeaths[getTeamNameFromTeamID(2)] = 0;
	$RocksTDMDeaths[getTeamNameFromTeamID(4)] = 0;
	$RocksTDMDeaths[getTeamNameFromTeamID(8)] = 0;

	EXTMissionStart();
}   


// --------------------------------------------------------------------------
// Displays an indication when a stealth name enters the game.
//
function RocksStealthOnAdd(%player)
{
	%name = getName(%player);

	if (strAlign(1, left, %name) == "/")
		say (0, 0, %name @ " joined the game.");
}   


// --------------------------------------------------------------------------
// Displays an indication when a stealth name exits the game.
//
function RocksStealthOnRemove(%player)
{
	%name = getName(%player);

	if (strAlign(1, left, %name) == "/")
		say (0, 0, %name @ " left the game.");
}   




// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// ------------------------- Player Functions -------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// Activates every time a player joins the game.  We log it, and show his IP
// in the Admin's chat window, assuming the Admin is in the game and has
// taken Admin control.
//
function RocksPlayerOnAdd(%player)
{
	%player.ipAdr = RocksGetAddressIP(%player, 4);

	%player.timeIn = getCurrentTime();

	%player.teamname = "";
	%player.monitor = getName(%player);
	%player.idle = $RocksIdleTime;

	%player.admin = 0;
	%player.selectIndex = 0;
	%player.selectPlayer = 0;
	%player.selectVehicle = 0;
	%player.pingStart = 0;
	%player.pingEnd = 0;
	%player.cheat = 0;

	RocksLogEntry("\tEntered" @ RocksInfoMaker(%player));


	if (RocksBanListCheck(%player))
		return;

	if (RocksR99Detector(%player))
		return;

	if (RocksTagDetector(%player))
		return;

	if (RocksUnkDetector(%player))
		return;

	if (!EXTPlayerOnAdd(%player))
		return;



	// we only do the MessageBox if the player's ClassA-IP is not
	// in our list of ClassA's, or if the check is not enabled
	//
	if ($RAKeyEnable) {
		if (!RAKeyCheck(%player)) {
			if ($RocksWelcomeMsg != "")
				messageBox(%player, $RocksWelcomeMsg);

			RocksTDMChoice(%player);
		}
	}
	else {
		if ($RocksWelcomeMsg != "")
			messageBox(%player, $RocksWelcomeMsg);

		RocksTDMChoice(%player);
	}


	RocksStealthOnAdd(%player);

	RocksShowPlayerInfo(0, %player, 1);

	RocksMissionStart(%player);


	// Start the Handle Monitor if it's not already running.  This covers
	// the case when the Monitor has been shut down because everyone
	// left, but then someone re-entered before the mission cycled.
	//
	if ((playerManager::getPlayerCount() == 1) &&
	    ($RocksR99Enable || $RocksTagEnable || ($RocksIdleTime > 0)))
		RocksMonitor(1);
}   


// --------------------------------------------------------------------------
// Activates every time a player leaves the game.
//
function RocksPlayerOnRemove(%player)
{
	if (isObject(%player)) {
		
		RocksStealthOnRemove(%player);

		%time = RocksET(%player.timeIn, getCurrentTime());


		// Bump the ET field out so it lines up
		//
		for (%i = strlen(%player.ipAdr); %i < 24; %i += 8)
			%time = "\t" @ %time;

		RocksLogEntry("\tExited" @ RocksInfoMaker(%player) @ %time);

		EXTPlayerOnRemove(%player);
	}
	else RocksLogEntry("\tExited\tDropped Connection");


	// Stop the Handle Monitor if the last player is leaving
	//
	if (playerManager::getPlayerCount() <= 1)
		RocksMonitor(0);
}   


// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// -------------------------- Vehicle Functions -----------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// Completely heal and reload a vehicle.  Note that the shields are not
// recharged.  The recharge of shields is controlled via the shield
// charge-rate in other files.  We don't wanna be messing with those files
// since that would modify the dynamics of combat.  I don't know why we have
// to do this so many times to achieve the heal.
//
function RocksVehicleHeal(%vehicle)
{
	if (!(%vehicle && isObject(%vehicle)))
		return;

	schedule("healObject(" @ %vehicle @ ", 50000);", 0.0);
	schedule("healObject(" @ %vehicle @ ", 50000);", 0.1);
	schedule("healObject(" @ %vehicle @ ", 50000);", 0.2);
	schedule("healObject(" @ %vehicle @ ", 50000);", 0.3);
	schedule("healObject(" @ %vehicle @ ", 50000);", 0.4);
	schedule("healObject(" @ %vehicle @ ", 50000);", 0.5);
	schedule("healObject(" @ %vehicle @ ", 50000);", 0.6);
	schedule("healObject(" @ %vehicle @ ", 50000);", 0.7);
	schedule("reloadObject(" @ %vehicle @ ", 50000);", 0.8);
}


//----------------------------------------------------------------------------
// Destroys a vehicle
//
function RocksVehicleDestroy(%vehicle)
{
	if (isObject(%vehicle)) {
		healobject(%vehicle, -50000);
		healobject(%vehicle, -50000);
	}
}


// --------------------------------------------------------------------------
// This function activates every time a vehicle is destroyed.  The goal of
// the function is to heal and reload the victor immediately after a fight.
// This helps reduce vulture-killing.  If RocksVehicleHeal() were called
// immediately, the player would still be 'on the trigger' at the end of a
// fight, so we delay the healing for an amount of time.  How MUCH time is
// up to you.
//
// The delay is controlled by setting the RocksHealDelay global at the top
// of this file.  It uses a global so that the value can be displayed in the
// game rules-tab.  Since it changes the dynamics of the game, it certainly
// should be explained in the rules-tab.
//
// I find that 2 seconds is good for virtually an immediate effect.  However,
// note that players are used to observing the amount of damage taken, and
// then hobbling to a heal-pad somewhere.  At the very least, they sometimes
// want to announce how badly damaged they are.  Therefore, I find a delay
// value of 8 seconds to be about right.
// 
function RocksVehicleOnDestroyed(%destroyed, %destroyer)
{
	// Reset the Idle Time.  Note that Admin are allowed to idle.
	//
	%player = playerManager::vehicleIdToPlayerNum(%destroyed);

	if (%player && !%player.admin)
		%player.idle = $RocksIdleTime;


	if (%destroyed != %destroyer) {
		$RocksTDMKills[getTeam(%destroyer)]++;
		$RocksTDMDeaths[getTeam(%destroyed)]++;
	}


	if (($RocksHealDelay > 0) && (%destroyed != %destroyer))
		schedule("RocksVehicleHeal (" @
			%destroyer @ ");", $RocksHealDelay);

	EXTVehicleOnDestroyed(%destroyed, %destroyer);
}


// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// ------------------------ TDM AutoTeam Management -------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// Initializes items used by the AutoTeam feature.
//
function RocksTDMInit(%init)
{
	$RocksTeamTag  = "";

	if ($server::TeamPlay == FALSE) {
		$RocksAutoTeam = FALSE;
		$RocksAutoTeamMode = 0;
		return;
	}

	// This block runs when the server first comes up
	//
	if (%init) {
		if (!$RocksAutoTeam)
			$RocksAutoTeamMode = 0;
		else $RocksAutoTeamMode = 1;

		return;
	}

	// This block runs whenever we change the
	// AutoTeam state via the Admin Control.
	//
	if ($RocksAutoTeam) {

		// Force Mode 1 if the team tag isn't set
		//
		if ($RocksAutoTeamTag == "") {
			$RocksAutoTeamMode = 1;
			return;
		}

		if ($RocksAutoTeamMode == 2)
			$RocksTeamTag = RocksStrToLower($RocksAutoTeamTag);
	}
}


// --------------------------------------------------------------------------
// Simply displays a message of which teams to choose from if it's TDM.
// If AutoTeam is on, then it will automatically choose the player's team.
//
function RocksTDMChoice(%player)
{

	if ($server::TeamPlay == false)
		return;


	if ($RocksAutoTeam) {
		if ($RocksTeamTag  != "")
			%txt = "AutoTeam (Tagged)";
		else %txt = "AutoTeam (UNTagged)";

		schedule("say(" @ %player @
			",0, \"" @ %txt @ " is on.\");", 2.0);
		return;
	}
}


// --------------------------------------------------------------------------
// Randomly chooses a team color from the list of available team colors.
// If the notTeam value is set, then it will choose a team that is NOT
// that color.  This is used for the AutoTeam feature.  If notTeam is 0,
// then the selection is from any of the allowed teams.
//
function RocksTDMRandomTeam(%notTeam)
{
	%count = 0;

	for (%i = 0; %i < 4; %i++) {
		%team = (1 << %i);

		if (%notTeam & %team)
			continue;

		if ($RocksAllowedTeams & %team) {
			%newteam[%count] = %team;
			%count++;
		}
	}

	return(%newteam[randomInt(0, %count-1)]);
}


// --------------------------------------------------------------------------
// Looks at all teams EXCEPT "notTeam", counts the number of players on each,
// subtracts one from the team that the vehicle is already on, and returns
// the team with the least number of players.  If "notTeam" is 0, then ALL
// teams are looked at.
//
function RocksTDMRouteTeam(%notTeam, %vehColor)
{
	// pick a random team if there's only one player in the server
	//
	if (playerManager::getPlayerCount() == 1)
		return(RocksTDMRandomTeam($RocksAutoTeamColor));


	// first find out which teams, and how many, we're selecting from
	//
	%count = 0;
	for (%i = 0; %i < 4; %i++) {
		%team = (1 << %i);

		if (%notTeam & %team)
			continue;

		if ($RocksAllowedTeams & %team) {
			%newteam[%count] = %team;
			%count++;
		}
	}

	// Now find the team with the least amount of players.
	// If all teams have equal players, then it doesn't matter.
	// We need to adjust for the team he's already on.
	//
	%min = $server::MaxPlayers;
	%max = 0;

	if (%vehColor == "Yellow Team")
		%colorID = 0x01;
	else if (%vehColor == "Blue Team")
		%colorID = 0x02;
	else if (%vehColor == "Red Team")
		%colorID = 0x04;
	else if (%vehColor == "Purple Team")
		%colorID = 0x08;
	else %colorID = 0;

	for (%i = 0; %i < %count; %i++) {

		%nPlayers = getNumberOfPlayersOnTeam(%newteam[%i]);

		if (%colorID == %newteam[%i])
			%nPlayers--;

		if (%nPlayers < %min) {
			%min = %nPlayers;
			%team = %i;
		}

		if (%nPlayers > %max)
			%max = %nPlayers;
	}

	// if all teams are equal, then let him change teams
	// as long as it's one of the allowed teams, otherwise
	// we pick it for him.
	//
	if (%min == %max) {
		if (!(%notTeam & %colorID)) {
			if ($RocksAllowedTeams & %colorID)
				return(%colorID);
		}
	}

	return(%newteam[%team]);
}


// --------------------------------------------------------------------------
// If AutoTeam is ON, this verifies that the player's team is one of the
// allowed teams.  This is done this way because the server doesn't
// enable/disable teams on a mission-basis.  Therefore, it's typical to enable
// all 4 teams.  When AutoTeam is ON, the Admin package will assure that only
// the desired teams are used, and that they are balanced.
// The player's own team-selection is ignored.
//
function RocksTDMBalance(%player, %vehicle)
{
	if (!$RocksAutoTeam)
		return;

	%vehColor = getTeam(%vehicle);

	// See block-comments below.
	//
//	if (%player.teamname != %vehColor) {

		// Choose from allowed teams EXCEPT the AutoTeamColor.
		// If the AutoTeam is Mode 2 and the player's name contains
		// the tag, then he's routed to the AutoTeamColor.  If we're
		// in Mode 1, then AutoTeamColor will be set to 0, which makes
		// it chooses from all available teams anyway.
		//
		%color = RocksTDMRouteTeam($RocksAutoTeamColor, %vehColor);


		// If we're in Mode 2, the player can always add the team-tag
		// to his name to get dropped on the AutoTeamColor
		//
		if ($RocksAutoTeamMode == 2) {
			%name = RocksStrToLower(getName(%player));

			if (RocksSubStr(%name, $RocksTeamTag))
				%color = $RocksAutoTeamColor;
		}

		%player.teamname = getTeamNameFromTeamID(%color);


	// Putting this if-statement here instead of up above, makes the
	// autoteam be fully automatic, i.e, it will ALWAYS keep the teams
	// balanced.  This means that when someone is destroyed, they might
	// respawn on another team if the teams have become unbalanced.
	// Leaving the if-statement here will cause it to rebalance only the
	// player actually tries to change teams, i.e. the teams could become
	// unbalanced if players from one team leave the server, yet no one
	// tries to change teams.
	//
	if (%player.teamname != %vehColor) {
		setTeam(%vehicle, %player.teamname);

		redrop(%vehicle);
	}
}


// --------------------------------------------------------------------------
// This assigns one of the allowed colors to the AutoTeamColor.
//
function RocksTDMColorSet(%new)
{
	// This assigns a random color to the AutoTeamColor if
	// it hasn't already been set by a TeamTag.
	//
	if (%new) {
		if ($RocksAutoTeamMode == 2)
			$RocksAutoTeamColor = RocksTDMRandomTeam(0);
		else $RocksAutoTeamColor = 0;
	}
	else {
		if ($RocksAutoTeamMode == 1)
			$RocksAutoTeamColor = 0;
		else if ($RocksAutoTeamMode == 2) {
			if ($RocksAutoTeamColor == 0)
				$RocksAutoTeamColor = RocksTDMRandomTeam(0);
		}
	}
}


// --------------------------------------------------------------------------
// Enables/Disables the AutoTeam feature from the Admin's Control
//
function RocksTDMAutoTeam(%admin, %mode)
{
	if ((%mode < 0) || (%mode > 2))
		return;

	$RocksAutoTeamMode = %mode;

	if ($RocksAutoTeamMode == 0) {
		$RocksAutoTeam = FALSE;

		RocksTDMInit(0);

		say(0,0,"\tAutoTeam has been deactivated");

		return;
	}

	$RocksAutoTeam = TRUE;

	RocksTDMInit(0);

	RocksTDMColorSet(0);


	// recheck the global value because the init function may
	// have reset it to mode 1.
	//
	if ($RocksAutoTeamMode == 1)
		say(0,0,"\tAutoTeam (UNTagged) has been activated.");
	else say(0,0,"\tAutoTeam (Tagged) has been activated.");


	// This drops everyone on the proper teams
	//
	%count = playerManager::getPlayerCount();

	for (%i = 0; %i < %count; %i++) {
		%player = playerManager::getPlayerNum(%i);

		%vehicle = playerManager::playerNumToVehicleId(%player);

		if (%vehicle) {
			%player.teamname = "";

			RocksTDMBalance(%player, %vehicle);
		}
	}
}


// --------------------------------------------------------------------------
function RocksTDMCtrlTimer(%admin, %mode)
{
	// assure the admin is still in the server
	//
	if (!isObject(%admin))
		return;

	%admin.tdmtimer -= 2;

	if (%admin.tdmtimer > 0) {
		schedule("RocksTDMCtrlTimer(" @ %admin @
				", " @ %mode @ ");", 2);
		return;
	}

	// only change the mode if it's different
	//
	if (%mode != $RocksAutoTeamMode)
		RocksTDMAutoTeam(%admin, $RocksAutoTeamMode);
}


// --------------------------------------------------------------------------
// Rotates thru the AutoTeam Modes of OFF, UNTAGGED, and TAGGED.
// The first call displays the current state, and subsequent calls bump it
// to the next state.
//
function RocksTDMAutoTeamCtrl(%admin)
{
	if ($server::TeamPlay == false) {
		say(%admin, 0, "\tServer is not in TDM mode.");
		return;
	}

	if (%admin.tdmtimer <= 0) {

		// Display the current status of the AutoTeam setting
		//
		if ($RocksAutoTeamMode == 0) {
			say(%admin, 0, "\tAutoTeam is currently OFF.");
		}
		else {
			if ($RocksAutoTeamMode == 1) {
				say(%admin, 0, "\tAutoTeam UNTagged " @
					"is currently enabled.");
			}
			else {
				say(%admin, 0, "\tAutoTeam Tagged " @
					"is currently enabled.");
			}
		}

		if ($RocksAutoTeamTag == "")
			say(%admin, 0, "There is no TeamTag set.");


		%admin.tdmtimer = 10;

		RocksTDMCtrlTimer(%admin, $RocksAutoTeamMode);

		return;
	}


	// We bump the TeamMode thru the various states and then
	// when the timer times out, it will set that mode.
	//
	%admin.tdmtimer = 10;

	$RocksAutoTeamMode++;


	// Can't run Mode2 if the TeamTag isn't set.
	//
	if (($RocksAutoTeamMode == 3) ||
	   (($RocksAutoTeamMode == 2) && ($RocksAutoTeamTag == "")))
		$RocksAutoTeamMode = 0;

	if ($RocksAutoTeamMode == 0)
		say(%admin, 0, "\tAutoTeam Off");
	else if ($RocksAutoTeamMode == 1)
		say(%admin, 0, "\tAutoTeam UNTagged On");
	else say(%admin, 0, "\tAutoTeam Tagged On");
}


// --------------------------------------------------------------------------
// This is someone else's script, which was hard to read, and not well-written.
// I just re-wrote it, cleaned it up, made it prettier, and adapted it for the
// Admin package.  I don't know who wrote the original.
//
//	124 - 133 Missiles and Arachs
//	147	  Aphids
//	134 - 136 Mines
//	150	  Sguns
//	110	  Plasmas
//
// Returns 1 if vehicle destruction was scheduled.
//
function RocksLimiter(%player, %vehicle, %limit)
{
	if (%limit <= 0)
		return(0);

	%count = 0;

	%wepCount = getWeaponCount(%vehicle); 

	for (%i = 0;  %i < %wepCount; %i++) {

		%id = getWeaponId(%vehicle, %i);

		if (((%id >= 124) && (%id <= 133)) ||
		     (%id == 147) || (%id == 110)  ||
		     (%id == 150) || ((%id >= 134) && (%id <= 136))) {
			%count++;
		}
	}


	if (%count > %limit) {

		say (%player, 0, "Missile/Mine/Sgun/Plasma Rack limit of " @
			%limit @ " exceeded.", "Rules_violated.wav");

		schedule("violate(" @ %vehicle @ ");", 1);

		schedule("RocksVehicleDestroy(" @ %vehicle @ ");", 2);

		return(1);
	}

	return(0);
}


// --------------------------------------------------------------------------
// Activates every time a player's vehicle, or an AI vehicle is spawned.
//
function RocksVehicleOnAdd(%vehicle)
{
	// This re-starts any custom Mission-based duties that were
	// turned off due to a mission end.  Otherwise, a player's
	// entry into the server does this
	//
	RocksMissionStart(0);

	%player = playerManager::vehicleIdToPlayerNum(%vehicle);

	if (%player) {

		if ($server::TeamPlay == TRUE)
			RocksTDMBalance(%player, %vehicle);

		if (%player.admin && ($server::TeamPlay == false))
			setTeam(%vehicle, $RocksAdminColor);

		%player.idle = 0;

		// Don't bother to continue if the limiter
		// is destroying the vehicle anyway.
		//
		if (RocksLimiter(%player, %vehicle, $RocksMMLimit))
			return;

		if ($RocksSuperScan)
			schedule("RocksCheatDetector (" @ %player @ ");", 3);
	}

	EXTVehicleOnAdd(%vehicle);
}


// --------------------------------------------------------------------------
// Activates whenever a player hits the Scan Key (I-Key) on another player in
// his target.  The game-default is to do nothing, however, if the player
// doing the scanning is Rock, then we might want to do something.
//
// With this function, you should hook in ONLY benign actions... if you place
// a Kick or Ban action in here, you might accidentally activate it in the
// heat of battle, and that would be bad.
//
function RocksVehicleOnScan(%scanned, %scanner)
{
	EXTVehicleOnScan(%scanned, %scanner);
}


// --------------------------------------------------------------------------
// Activates whenever a player hits the Scan Key (I-Key) on a structure.
// Again, the game-default is to do nothing, and again only benign actions.
//
function RocksStructureOnScan(%scanned, %scanner)
{
	EXTStructureOnScan(%scanned, %scanner);
}


// --------------------------------------------------------------------------
// This function activates whenever a vehicle targets another vehicle.  The
// game-default is to do nothing.  However, the Admin may want to do something
// with this hook.  So far though, I can't think of anything useful to do.
//
function RocksVehicleOnTargeted(%targeted, %targeter)
{
	EXTVehicleOnTargeted(%targeted, %targeter);
}


// --------------------------------------------------------------------------
// Activates whenever a vehicle shoots another
//
function RocksVehicleOnAttacked(%attackee, %attacker)
{
	EXTVehicleOnAttacked(%attackee, %attacker);
}


// --------------------------------------------------------------------------
// Activates whenever a vehicle arrives at a location.
//
function RocksVehicleOnArrived(%vehicle, %where)
{
	EXTVehicleOnArrived(%vehicle, %where);
}


// --------------------------------------------------------------------------
// Sends out loud and obnoxious warnings that the server is going to restart.
// Timelimit must be at least 5 seconds.
//
function RocksRestartWarning(%timelimit)
{
	if (%timelimit < 5)
		return;

	if (%timelimit >= 60)
		schedule("say(0, 0, \"\tWARNING: Server restart in " @
			"60 Seconds\");", %timelimit - 60);

	if (%timelimit > 30)
		schedule("say(0, 0, \"\tWARNING: Server restart in " @
			"30 Seconds\");", %timelimit - 30);

	if (%timelimit > 10) {
		schedule("say(0, 1, \"\", \"sfx_siren.wav\");",
			%timelimit - 10);

		schedule("say(0, 0, \"\tWARNING: Server restart in " @
			"10 Seconds\");", %timelimit - 10);
	}

	schedule("say(0, 2, \"\", \"sfx_siren.wav\");", %timelimit - 4.5);

	schedule("say(0, 0, \"\tWARNING: Server restart in 5 Seconds\");",
		%timelimit - 5);

	schedule("say(0, 0, \"\tWARNING: Server restart in 4...\");",
		%timelimit - 4);

	schedule("say(0, 0, \"\tWARNING: Server restart in 3...\");",
		%timelimit - 3);

	schedule("say(0, 0, \"\tWARNING: Server restart in 2...\");",
		%timelimit - 2);

	schedule("say(0, 0, \"\tWARNING: Server restart in 1...\");",
		%timelimit - 1);

	schedule("messageBox(0, \"Server restarting.\\n\\nPlease rejoin.\");",
		%timelimit - 1);
}


// --------------------------------------------------------------------------
// Writes the autoconfig file
//
function RocksWriteAutoConfig(%skirmish, %teams, %teamcolors,
			%nplayers, %password, %mission, %hostname,
			%timelimit, %allowances, %autoheal)
{
	export("$RocksAutoConfigFile", "RocksAutoConfig.cs", overwrite);

	fileWrite($RocksAutoConfigFile, overwrite,
		"// This is an automatically generated file.  DO NOT EDIT!\n");

	fileWrite($RocksAutoConfigFile, append,
		"$RocksSkirmishGame		= " @ %skirmish @ ";");

	fileWrite($RocksAutoConfigFile, append,
		"$RocksSkirmishTeams		= " @ %teams @ ";");

	fileWrite($RocksAutoConfigFile, append,
		"$RocksSkirmishTeamColors	= " @ %teamcolors @ ";");

	fileWrite($RocksAutoConfigFile, append,
		"$RocksSkirmishNplayers		= " @ %nplayers @ ";");

	fileWrite($RocksAutoConfigFile, append,
		"$RocksSkirmishPassword		= \"" @ %password @ "\";");

	fileWrite($RocksAutoConfigFile, append,
		"$RocksSkirmishMission		= \"" @ %mission @ "\";");

	fileWrite($RocksAutoConfigFile, append,
		"$RocksSkirmishHostName		= \"" @ %hostname @ "\";");

	fileWrite($RocksAutoConfigFile, append,
		"$RocksSkirmishTimeLimit	= " @ %timelimit @ ";");

	fileWrite($RocksAutoConfigFile, append,
		"$RocksSkirmishAllowances	= " @ %allowances @ ";");

	fileWrite($RocksAutoConfigFile, append,
		"$RocksSkirmishAutoheal		= " @ %autoheal @ ";");
}


// --------------------------------------------------------------------------
// Reboots the Server with a new configuration controlled by the remote
//
function RocksServerRestart(%player, %skirmish, %teams, %teamcolors,
			%nplayers, %password, %mission, %hostname,
			%timelimit, %allowances, %autoheal)
{
	if ($CmdLineServer != TRUE) {
		say(%player, 0, "\tInvalid on Non-Dedicated Servers");
		return;
	}

	if ($RocksServerCycle > 1) {
		say(%player, 0, "\tAlready in Server-Restart mode");
		return;
	}

	if (!EXTServerRestart (%player, %skirmish, %teams, %teamcolors,
			%nplayers, %password, %mission, %hostname,
			%timelimit, %allowances, %autoheal)) {
		say(%player, 0, "\tServer restart aborted");
		return;
	}

	if ($RocksAutoConfigFile == "") {
		say(%player, 0, "\tInvalid/corrupt Auto Config Filename");
		return;
	}

	// First check that all the parametes make sense
	//
	if (!((%skirmish == 1) || (%skirmish == true)
	      || (%skirmish == 0) || (%skirmish == false))) {
		say(%admin, 0, "\tInvalid SkirmishGame field");
		return;
	}

	if (%skirmish) {
		if (!isFile($RocksServerFlipperFile)) {
			say(%admin, 0, "\tCannot find file " @
					$RocksServerFlipperFile);
			return;
		}

		if ((%teams < 1) || (%teams > 4)) {
			say(%admin, 0, "\tInvalid SkirmishTeams field");
			return;
		}

		if (%teams > 1) {
			if ((%teamcolors < 1) || (%teamcolors > 1111)) {
				say(%admin, 0,
					"\tInvalid SkirmishTeamColors field");
				return;
			}
		}

		if ((%nplayers < 1) || (%nplayers > 32)) {
			say(%admin, 0, "\tInvalid SkirmishNplayers field");
			return;
		}

		if (!isFile("multiplayer\\\\" @ %mission @ ".cs")) {
			say(%admin, 0, "\tInvalid SkirmishMission field");
			return;
		}

		if ((%allowances < 1) || (%allowances > 111111)) {
			say(%admin, 0, "\tInvalid SkirmishAllowances field");
			return;
		}

		// autoheal can be a negative number (off, yet set)
	}
	else {
		%hostname = "";
		%password = "";
		%mission = "";
	}


	RocksWriteAutoConfig(%skirmish, %teams, %teamcolors,
			%nplayers, %password, %mission, %hostname,
			%timelimit, %allowances, %autoheal);


	// make sure we haven't already entered into
	// the end-of-maplist restart sequence.
	//
	if ($RocksServerCycle != 1) {
		RocksRestartWarning(60);
		schedule("quit();", 60);
	}
	else say(%player, 0, "\tServer Restarting... already sequencing.");


	// prevent function re-entry until the new game starts
	//
	$RocksServerCycle = 2;
}


// --------------------------------------------------------------------------
// Cycle to the next mission-map
//
function RocksGameRestart(%player)
{
	if ($RocksGameCycle == FALSE) {
		say(%player, 0, "\tAlready in Game-Cycle mode");
		return;
	}

	// prevent function re-entry until the new game starts
	//
	$RocksGameCycle = FALSE;


	if ($CmdLineServer == TRUE) {

		if ($RocksMapNumber == $RocksMapCount)
			%mapN = 0;
		else %mapN = $RocksMapNumber;

		eval("%map = $MissionCycling::Stage" @ %mapN @ ";");

		say(0, 0, "\tCycling to " @ %map @  " (" @ %mapN+1 @ " of "
				@ $RocksMapCount @ ") in 10 seconds.");
	}
	else say(0, 0, "\tEnding current map in 10 seconds.");


	schedule("missionEndConditionMet();", 10);
}


// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// ---------------------------- Rules Tab -----------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// Displays the rules in the Rules Tab.  Note that the length of the rules
// is limited.  Anything longer than this and the game will chop it off
// when it is displayed.
//
function RocksRules()
{
	%rules = EXTRules();
	if (%rules != "" ) {
		setGameInfo(%rules);      
		return;
	}


	if($server::TeamPlay == true) {
		%rules ="<tIDMULT_TDM_GAMETYPE>"   @        
			"<tIDMULT_TDM_MAPNAME>"    @ 
			$missionName               @
			"<tIDMULT_TDM_OBJECTIVES>" @
			"<tIDMULT_TDM_SCORING_1>"  @
			"<tIDMULT_TDM_SCORING_2>"  @
			$killPoints                @
			"<tIDMULT_TDM_SCORING_3>"  @
			"<tIDMULT_TDM_SCORING_4>"  @
			$deathPoints               @
			"<tIDMULT_TDM_SCORING_5>"  @
			"<tIDMULT_TDM_SCORING_6>"  @
			"<tIDMULT_STD_ITEMS>"      @
			"<tIDMULT_STD_HEAL>"       @
			"<tIDMULT_STD_RELOAD_1>"   @
			$PadWaitTime               @
			"<tIDMULT_STD_RELOAD_2>"   @
			"<tIDMULT_STD_ZEN_1>"      @
			$ZenWaitTime               @
			"<tIDMULT_STD_ZEN_2>";
	}
	else {
		%rules ="<tIDMULT_DM_GAMETYPE>"    @        
			"<tIDMULT_DM_MAPNAME>"     @ 
			$missionName               @
			"<tIDMULT_DM_OBJECTIVES>"  @
			"<tIDMULT_DM_SCORING_1>"   @
			"<tIDMULT_DM_SCORING_2>"   @
			$killPoints                @
			"<tIDMULT_DM_SCORING_3>"   @
			"<tIDMULT_DM_SCORING_4>"   @
			$deathPoints               @
			"<tIDMULT_DM_SCORING_5>"   @
			"<tIDMULT_STD_ITEMS>"      @
			"<tIDMULT_STD_HEAL>"       @
			"<tIDMULT_STD_RELOAD_1>"   @
			$PadWaitTime               @
			"<tIDMULT_STD_RELOAD_2>"   @
			"<tIDMULT_STD_ZEN_1>"      @
			$ZenWaitTime               @
			"<tIDMULT_STD_ZEN_2>";
                  
	}

	setGameInfo(%rules);      
}




// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// ---------------------------- Game Log File -------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// Initializes Rock's Game Log.  This assures we don't run Rock's Log at
// the same time as Starsiege's built-in log... we don't need both of them.
// This also assures we don't run ANY log if the RocksAdminLog value is FALSE.
//
function RocksLogInit()
{
	if ($RocksAdminLog == FALSE) {
		$missionLogFile = "";
		return;
	}

	if ($RocksSSLog == TRUE) {
		$missionLogFile = $RocksLogFilename;
		$RocksAdminLog = FALSE;
		return;
	}
	else $missionLogFile = "";


	// Put a blank line in Rock's Game Log if it's being used
	//
	if ($RocksLogFilename != "") {
		fileWrite($RocksLogFilename, append, "");
		RocksLogEntry("Log Opened");
	}
	else $RocksAdminLog = FALSE;
}


// --------------------------------------------------------------------------
// Logs strings to Rock's Game Log if it's enabled
//
function RocksLogEntry(%logstring)
{
	if ($RocksAdminLog)
		fileWrite($RocksLogFilename, append,
			getDate(), " ", getTime(), " - ", %logstring);
}




// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// ------------------------ IP Ban List Management --------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// Reads the IP Ban List file into memory.
//
function RocksBanListInit()
{
	$BanCount = 0;

	if ($RocksBanFilename == "")
		return;

	if (!isFile($RocksBanFilename)) {

		// No File, so create it by writing some formatting info
		//
		fileWrite($RocksBanFilename, append, "");

		fileWrite($RocksBanFilename, append,
			"// Each Banned IP contains 3 lines.",
			"  Their format is:");

		fileWrite($RocksBanFilename, append, "//");

		fileWrite($RocksBanFilename, append,
			"//	$BanName[$BanCount]	= \"BadBoy\";");

		fileWrite($RocksBanFilename, append,
			"//	$BanIP[$BanCount]	=",
			" \"IP:192.168.1.100:0\";");

		fileWrite($RocksBanFilename, append, "//	$BanCount++;");

		fileWrite($RocksBanFilename, append, "//");

		fileWrite($RocksBanFilename, append, "");

		return;
	}

	exec($RocksBanFilename);
}


// --------------------------------------------------------------------------
// This is used for maintenance admin, i.e. admin-activity that does not fall
// within the more serious catagory of disciplinary actions.
//
function RocksLittleFoot(%player, %code)
{
	%info = RocksInfoMaker(%player);

	%saystr = " AutoKicked - " @
		getName(%player) @ "  " @ %player.ipAdr;

	if (%code == 1) {
		%logstr = "\tCTags" @ %info;

		%saystr =  "CTag Detection" @ %saystr;

		kick (%player, "\n\nImproper angle-brackets." @
				"\nOnly <F1> thru <F8> are permitted.");
	}

	if (%code == 2) {
		%logstr = "\tIdle" @ %info;

		%saystr =  "Idle Detection" @ %saystr;

		kick (%player, "\n\nInactivity Auto-kick");
	}

	if (%code == 3) {
		%logstr = "\tUnk" @ %info;

		%saystr =  "Unk Detection" @ %saystr;

		kick (%player, "\n\nYou need a REAL game-name " @
				"before you can enter.");
	}

	RocksLogEntry(%logstr);

	say(0, 0, %saystr);
}


// --------------------------------------------------------------------------
// This executes on a schedule to ban a player who's IP address was listed
// in the Ban List, or a player who attempts to use the R99-crasher algorithm.
// The player is banned for 4 hours.  We have to schedule the action because
// it happens so quickly that any entry-message may not be seen by the
// offending player... and this is one message we definitely want him to see.
// If just the 'ban' was scheduled instead of this whole function, then
// the logging and string-display would be out-of-sync with the actual event,
// then this would be an incorrect algorithm.
//
function RocksBigBoot(%player, %index, %which)
{
	%info = RocksInfoMaker(%player);

	%saystr = " AutoBanned - " @
			getName(%player) @ "  " @ %player.ipAdr;

	if (%which == 0) {
		%saystr = "BanList Entry" @ %saystr @
				" (" @ $BanName[%index] @ ")";

		%logstr = "\tAutoBan" @ %info @ " (" @ $BanName[%index] @ ")";
	}
	else if (%which == 1) {
		%saystr =  "R99 Detection" @ %saystr;

		%logstr = "\tR99Ban" @ %info;
	}
	else if (%which == 2) {
		%saystr =  "CheatRig Detection" @ %saystr;

		%logstr = "\tRigBan" @ %info;
	}

	ban(%player, 240,
		"\nYou have been PERMANENTLY AutoBanned.\n\n");

	RocksLogEntry(%logstr);

	say(0, 0, %saystr);
}


// --------------------------------------------------------------------------
// This is once-removed from RocksBigBoot so that we can display a larger
// message to the player.  The ban-box message has a length limitation, so
// this let's us put a longer message up.  I use it to give info on how the
// player can protest his inclusion in the banlist.
//
function RocksBigFoot(%player, %index, %which)
{
	schedule("RocksBigBoot (" @ %player @ ", " @
				%index @ ", " @ %which @ ");", 0.5);

	if ($RocksAdminEmail != "") {
		messageBox (%player, "\nYou have been PERMANENTLY AutoBanned.\n\nYou may protest and be reconsidered for entry into this server by emailing your game-name, date, and approximate time of this event to:\n\n" @ $RocksAdminEmail);
	}
}


// --------------------------------------------------------------------------
// This function gets called every time a new player enters the game.  It
// checks the player's IP against those in the IP Ban List.  If a match is
// found, the player is banned with RocksBigFoot(), unless the player's IP
// is entered into the Ban List as a pass-thru entry.
// Returns 1 if a match was found, else returns 0 if the player was not
// in the list, or should not be banned.
//
function RocksBanListCheck(%player)
{
	%name = RocksStrToLower(getName(%player));

	%subnet = RocksGetAddressIP(%player, 3);

	for(%i = 0; %i < $BanCount; %i++) {

		%flag = strAlign(4, left, $BanIP[%i]);

		if (%flag == "PIP:") {		// Pass: let this IP in

			%majnet = RocksStrRight($BanIP[%i],
					strlen($BanIP[%i]) - 1);

			if (isEqualIP(%majnet, %subnet) ||
			    isEqualIP(%subnet, %majnet)) {
				return(0);
			}
		}
		else if (%flag == "SIP:") {	// Sub: sub-IP match

			%majnet = strAlign(strlen($BanIP[%i]),
					left, "S" @ %subnet);

			if ($BanIP[%i] == %majnet) {
				RocksBigFoot (%player, %i, 0);
				return(1);
			}
		}
		else if (%flag == "BIP:") {	// Both: Name AND IP must match

			if (RocksSubStr(%name, RocksStrToLower($BanName[%i]))) {
				%majnet = strAlign(strlen($BanIP[%i]),
						left, "B" @ %subnet);

				if ($BanIP[%i] == %majnet) {
					RocksBigFoot (%player, %i, 0);
					return(1);
				}
			}
		}
		else if (%flag == "NIP:") {	// Name: name-only

			if (RocksSubStr(%name, RocksStrToLower($BanName[%i]))) {
				RocksBigFoot (%player, %i, 0);
				return(1);
			}
		}
		else if (isEqualIP($BanIP[%i], %subnet) ||
			 isEqualIP(%subnet, $BanIP[%i])) {
				// General IP-ban
				//
				RocksBigFoot (%player, %i, 0);
				return(1);
		}
	}

	return(0);
}


// --------------------------------------------------------------------------
// This is used to write a player's IP address, in the proper format, to the
// Ban List file.  It doesn't ban the player, it just adds him to the file.
//
function RocksBanListAdd(%player)
{
	if (!isFile($RocksBanFilename)) {
		RocksAnnounce("\tBanList File not found.");
		return;
	}


	// Update our in-memory list
	//
	%index = $BanCount;

	$BanName[$BanCount] = getName(%player);
	$BanIP[$BanCount] = %player.ipAdr;
	$BanCount++;


	// Now write it out to file
	//
	fileWrite($RocksBanFilename, append,
		"$BanName[$BanCount]\t= \"", $BanName[%index], "\";");

	fileWrite($RocksBanFilename, append,
		"$BanIP[$BanCount]\t= \"", $BanIP[%index], "\";");

	fileWrite($RocksBanFilename, append, "$BanCount++;");

	fileWrite($RocksBanFilename, append, "");


	%str = "\tBlisted" @ RocksInfoMaker(%player);

	RocksLogEntry(%str);

	RocksAnnounce(%str);
}




// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// -------------------------- Tag Detector ----------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// Searches a player's name/handle for appropriate color tags.  Any angle
// brackets that are found without color tags will result in a kick.
// Returns 1 if bad tags are detected.
//
function RocksTagDetector(%player)
{
	if (!$RocksTagEnable)
		return(0);

	%name = getName(%player);

	%count = RocksSubStr(%name, "<");

	if (!%count)
		return(0);

	%count1 = 0;
	%count2 = 0;

	for (%i = 1; %i < 9; %i++) {
		%count1 += RocksSubStr(%name, "<f" @ %i @ ">");
		%count2 += RocksSubStr(%name, "<F" @ %i @ ">");
	}

	if (%count != (%count1 + %count2)) {
		schedule("RocksLittleFoot (" @ %player @ ", 1);", 0.5);
		return(1);
	}

	return(0);
}



// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// ------------------------ Unknown Detector --------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// Ejects a player with an "unknown" handle.
// Returns 1 if detected and ejected.
//
function RocksUnkDetector(%player)
{
	if (!$RocksNoUnkEnable)
		return(0);

	%name = RocksStrToLower(getName(%player));

	if (RocksSubStr(%name, "unknown")) {
		schedule("RocksLittleFoot (" @ %player @ ", 3);", 0.5);
		return(1);
	}

//	if (RocksSubStr(%name, "trainee")) {
//		schedule("RocksLittleFoot (" @ %player @ ", 3);", 0.5);
//		return(1);
//	}

	return(0);
}



// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// -------------------------- Idle Detector ---------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// Ejects a player for inactivity.  He gets a 60-second warning first.
//
function RocksIdleDetector(%player, %time)
{
	// This checks for the window represented by the monitor rate
	//
	if ((%time >= (60 - ($RocksMonitorRate - 1))) && (%time <= 60)) {
		messageBox(%player, "Inactivity Warning\n\n" @
			"You will be Auto-Kicked in 60 seconds");
		return;
	}

	if (!%time)
		schedule("RocksLittleFoot (" @ %player @ ", 2);", 0.5);
}



// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// -------------------------- Cheat Detector --------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// AutoBans a player if the SuperScanner detects an invalid vehicle.
//
function RocksCheatDetector(%player)
{
	if (!isObject(%player))
		return;


	// this is set to 0 when the player enters the server, and it gets
	// set to 1 by the SuperScanner if it detects an invalid vehicle.
	//
	if (!%player.cheat)
		return;


	RocksBanListAdd(%player);

	RocksBigFoot (%player, 0, 2);
}



// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// ------------------------- R99 Detector -----------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// Searches a player's name/handle for the r99 strings and, if a match is
// found, automatically adds him to the permanent Ban List, then bans him
// and notifies all players of the event and broadcasts his IP.
// Returns 1 if detected.
//
$RocksR99Count		= 0;
$RocksR99String[$RocksR99Count]	= "<r"; $RocksR99Count++;
$RocksR99String[$RocksR99Count]	= "<R"; $RocksR99Count++;
$RocksR99String[$RocksR99Count]	= "<y"; $RocksR99Count++;
$RocksR99String[$RocksR99Count]	= "<Y"; $RocksR99Count++;
$RocksR99String[$RocksR99Count]	= "<9"; $RocksR99Count++;
$RocksR99String[$RocksR99Count]	= "99>"; $RocksR99Count++;

function RocksR99Detector(%player)
{
	if (!$RocksR99Enable)
		return(0);

	%name = getName(%player);

	for(%i = 0; %i < $RocksR99Count; %i++) {
		if (RocksSubStr(%name, $RocksR99String[%i])) {
			RocksBanListAdd(%player);
			RocksBigFoot (%player, 0, 1);
			return(1);
		}
	}

	return(0);
}


// --------------------------------------------------------------------------
// This is a generic player monitor that is used to check players for things
// like inappropriate color-tags, R99 crash attempts, game inactivity, etc.
// The "enabled" parameter is used to turn it on and off.
//
$RocksMonitorOnoff = 0;
$RocksMonitorStart = 0;
$RocksMonitorPending = 0;

function RocksMonitor(%enabled)
{
	// Prevent function re-entry
	//
	if (!%enabled) {
		$RocksMonitorStart = 0;
		$RocksMonitorOnoff = 0;
		$RocksMonitorPending = 0;
		return;
	}
	else if ((%enabled == 1) && ($RocksMonitorRate)) {
		if ($RocksMonitorStart)
			return;

		$RocksMonitorStart = 1;
		$RocksMonitorOnoff = 1;

		if ($RocksMonitorPending)
			return;
	}
	else if ($RocksMonitorOnoff == 0) {
		$RocksMonitorPending = 0;
		return;
	}


	echo("Rock's Monitor");

	$RocksMonitorPending = 1;

	%count = playerManager::getPlayerCount();

	for (%i = 0; %i < %count; %i++) {
		%player = playerManager::getPlayerNum(%i);

		if (%player) {
			%name = getName(%player);

			if (%player.monitor != %name) {

				%player.monitor = %name;

				RocksR99Detector(%player);

				RocksTagDetector(%player);
			}


			if (%player.idle > 0) {
				%player.idle -= $RocksMonitorRate;

				// adjust for overshoot
				//
				if (%player.idle < 0)
					%player.idle = 0;

				RocksIdleDetector(%player, %player.idle);
			}
		}
	}

	schedule("RocksMonitor(2);", $RocksMonitorRate);
}




// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// ------------------------- Remote Control ---------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// This is where most Administrator functions are performed.  This remote
// control function lets the Admin perform admin duties from his own gameplay
// computer, and he doesn't have to be sitting at the server-machine.  The
// methodology that permits this really cool feature is Keybinding and
// Starsiege's remoteEval() function.
//
// Keybinding can be confusing.  Keybinding takes place on the client side
// of the game.  In other words, keybinding is performed by the game players,
// not the game server.  That's why you don't find the binding in this file.
// To operate these remote control features in this section, the Admin's
// keymap file in the ./keyMaps directory, ROCK.CS in my case, needs to have
// a few keys bound such that these remote-control functions will operate.
//
// The keymap file is NOT the one on the server-machine... it's the one on
// the machine that the Admin is playing the game on, unless of course they are
// the same machine.  It is possible, and common, to start a dedicated server,
// leave it running on the desktop, and then go fire up and play the Starsiege
// game in the normal manner.
//
// Anyways, in the Admin's keymap file will have lines like the following:
// 
//   bind(keyboard0, make, alt, "w", TO, "remoteEval(2048,RocksRemote,1);");
//   bind(keyboard0, make, alt, "k", TO, "remoteEval(2048,RocksRemote,2);");
//   bind(keyboard0, make, alt, "b", TO, "remoteEval(2048,RocksRemote,3);");
//
// For example, the last binding calls remoteEval() when ALT-b is pressed
// on the Admin's keyboard.
//
// Here's how remoteEval() works.  The remoteEval() function needs at least
// two parameters.  The first one needs to be the number 2048.  This number
// represents Starsiege's managerID.  I don't know why it's 2048 or if there
// are other numbers that work, but if you haven't yet noticed, the very first
// player to enter a server is always 2049.
//
// The second parameter is a word that, when concatenated onto the end of the
// word 'remote' forms the function-name, i.e. 'remoteRocksRemote'.  Are you
// still with me so far?  Now the resulting function, remoteRocksRemote(),
// is passed the playerID of the player that made the function-call (via his
// bound keystroke) along with N additional parameters.  The max of N is
// unknown, but suffice it to say that one or two additional parameters are
// plenty for our use.
//
// That last parameter (and we could have more if we wanted) is simply
// a number that Rock uses to indicate to the remote function (and therefore
// the server) what is to be done.  From this, you can code up all sorts of
// Administrator functions.
//
function remoteRocksRemote(%player, %code,
			%code1, %code2, %code3, %code4, %code5,
			%code6, %code7, %code8, %code9, %code10)
{

	if (%code == $AdminPassword) {
		RocksControlGrabber(%player);

		%player.selectPlayer = 0;
		%player.selectVehicle = 0;

		return;
	}


	// Assure that only Admin performs these activities
	//
	if (!%player.admin) {
		say (%player, 0, "\tYou do not have Admin Access.");

		RocksAnnounce("\t" @ getName(%player) @
				" attempted to gain Admin Control.");
		return;
	}


	%retval = EXTcode(%player, %player.selectPlayer,
			%player.selectVehicle, %code,
			%code1, %code2, %code3, %code4, %code5,
			%code6, %code7, %code8, %code9, %code10);

	if (%retval != 0) {
		// make sure the admin is still here
		//
		if (isObject(%player)) {
			if (%retval == 1)
				%player.selectPlayer = 0;
			else if (%retval == 2)
				%player.selectVehicle = 0;
			else if (%retval == 3) {
				%player.selectPlayer = 0;
				%player.selectVehicle = 0;
			}
		}

		return;
	}



	// -------------------------------------------------------------
	// -------------------------------------------------------------
	// These codes don't need a selected player to operate
	//
	if (%code == 21) {			// Cycle Game
		RocksGameRestart(%player);
		return;
	}


	if (%code == 22) {			// Auto-Heal on/off
		$RocksHealDelay = -$RocksHealDelay;

		if ($RocksHealDelay < 0)
			%str = "\tAuto-Heal Disabled";
		else if ($RocksHealDelay > 0)
			%str = "\tAuto-Heal Enabled";
		else %str = "\tAuto-Heal is Permanently Off";

		// Update the Rules Tab
		//
		RocksRules();
		RocksAnnounce(%str);
		return;
	}


	if (%code == 23) {			// Game Log on/off
		if ($RocksAdminLog == TRUE) {
			RocksLogEntry("Log Closed");
			$RocksAdminLog = FALSE;
			%str = "\tGame Log Disabled";
		}
		else {
			$RocksAdminLog = TRUE;
			%str = "\tGame Log Enabled";
		}

		// Resync the Game Log
		//
		RocksLogInit();
		RocksAnnounce(%str);
		return;
	}


	if (%code == 24) {			// Show all players
		RocksShowPlayerInfo(%player, 0, 0);
		return;
	}


	if (%code == 25) {			// Select player forward
		RocksPlayerSelect(%player, 1);
		return;
	}


	if (%code == 26) {			// Select player backward
		RocksPlayerSelect(%player, 0);
		return;
	}


	if (%code == 27) {			// ATR scoring on/off
		if ($RocksATRscoring) {
			$RocksATRscoring = FALSE;
			%str = "\tATR scoring Disabled";
		}
		else {
			$RocksATRscoring = TRUE;
			%str = "\tATR scoring Enabled";
		}

		RocksRules();
		RocksAnnounce(%str);
		return;
	}


	if (%code == 28) {			// R99 Monitor on/off
		if ($RocksR99Enable) {
			$RocksR99Enable = FALSE;
			%str = "\tR99 Monitor Disabled";
		}
		else {
			$RocksR99Enable = TRUE;
			%str = "\tR99 Monitor Enabled";
		}

		RocksAnnounce(%str);

		if (($RocksR99Enable == FALSE) &&
		    ($RocksTagEnable == FALSE) &&
		    ($RocksIdleTime <= 0))
			RocksMonitor(0);
		else RocksMonitor(1);

		return;
	}


	if (%code == 29) {			// Select the targeted vehicle
		%player.selectPlayer = RocksTargetedPlayer(%player);

		if (%player.selectPlayer)
			RocksShowPlayerInfo(%player, %player.selectPlayer, 1);
		else {
			%player.selectVehicle = RocksTargetedVehicle(%player);

			if (%player.selectVehicle) {
				say (%player, 0, "\t" @
					getHUDName(%player.selectVehicle) @
					" (Selected)");
			}
			else say (%player, 0, "\tNo Targeted Vehicle");
		}

		return;
	}


	if (%code == 30) {			// Select next Music group
		RocksCDAudio(%player, 1);
		return;
	}


	if (%code == 31) {			// Music Off
		RocksCDAudio(%player, 0);
		return;
	}


	if (%code == 32) {			// Restart Server Default
		RocksServerRestart(%player, 0, 0, 0, 0, "", "", "", 0, 0, 0);
		return;
	}


	if (%code == 33) {			// Restart Server Skirmish
		RocksServerRestart(%player, %code1, %code2, %code3, %code4,
			%code5, %code6, %code7, %code8, %code9, %code10);
		return;
	}


	if (%code == 34) {			// Tag Monitor on/off
		if ($RocksTagEnable) {
			$RocksTagEnable = FALSE;
			%str = "\tTag Monitor Disabled";
		}
		else {
			$RocksTagEnable = TRUE;
			%str = "\tTag Monitor Enabled";
		}

		RocksAnnounce(%str);

		if (($RocksR99Enable == FALSE) &&
		    ($RocksTagEnable == FALSE) &&
		    ($RocksIdleTime <= 0))
			RocksMonitor(0);
		else RocksMonitor(1);

		return;
	}


	if (%code == 35) {			// Idle Monitor on/off
		$RocksIdleTime = -$RocksIdleTime;

		if ($RocksIdleTime < 0)
			%str = "\tIdle Monitor Disabled";
		else if ($RocksIdleTime > 0)
			%str = "\tIdle Monitor Enabled";
		else %str = "\tIdle Monitor is Permanently Off";


		%count = playerManager::getPlayerCount();

		for (%i = 0; %i < %count; %i++) {
			%player = playerManager::getPlayerNum(%i);

			if (%player.admin ||
			    playerManager::playerNumToVehicleId(%player))
				%player.idle = 0;
			else %player.idle = $RocksIdleTime;
		}


		RocksAnnounce(%str);

		if (($RocksR99Enable == FALSE) &&
		    ($RocksTagEnable == FALSE) &&
		    ($RocksIdleTime <= 0))
			RocksMonitor(0);
		else RocksMonitor(1);

		return;
	}


	if (%code == 36) {			// NoUnk on/off
		if ($RocksNoUnkEnable) {
			$RocksNoUnkEnable = FALSE;
			%str = "\tNoUnk Disabled";
		}
		else {
			$RocksNoUnkEnable = TRUE;
			%str = "\tNoUnk Enabled";
		}

		RocksAnnounce(%str);
		return;
	}


	if (%code == 38) {			// Random Thoughts on/off
		if ($RocksRandomThoughts) {
			RRTStop();
			$RocksRandomThoughts = FALSE;
			%str = "\tRandomThoughts Disabled";
		}
		else {
			RRTStart();
			$RocksRandomThoughts = TRUE;
			%str = "\tRandomThoughts Enabled";
		}

		RocksAnnounce(%str);
		return;
	}


	if (%code == 39) {			// AutoTeam Control
		RocksTDMAutoTeamCtrl(%player);
		return;
	}


	if (%code == 40) {			// Missile/Mine Limiter on/off
		$RocksMMLimit = -$RocksMMLimit;

		if ($RocksMMLimit < 0)
			%str = "\tMissileMine Limiter Disabled";
		else if ($RocksMMLimit > 0)
			%str = "\tMissileMine Limiter Enabled";
		else %str = "\tMissileMine Limiter is Permanently Off";

		RocksAnnounce(%str);
		return;
	}



	// This if-statement wraps these codes because the Admin may have
	// selected a clone vehicle which has no player, or the player that
	// he has selected is no longer in the game.
	//
	if (%player.selectPlayer && isObject(%player.selectPlayer)) {

		if (%code == 1) {		// Warn
			RocksFoot(%player, %player.selectPlayer, 1);
			return;
		}


		if (%code == 2) {		// Kick
			RocksFoot(%player, %player.selectPlayer, 2);

			// make sure the admin is still here
			//
			if (isObject(%player)) {
				%player.selectPlayer = 0;
				%player.selectVehicle = 0;
			}

			return;
		}


		if (%code == 3) {		// Ban
			RocksFoot(%player, %player.selectPlayer, 3);

			// make sure the admin is still here
			//
			if (isObject(%player)) {
				%player.selectPlayer = 0;
				%player.selectVehicle = 0;
			}

			return;
		}


		if (%code == 4) {		// Reassign Admin
			if (%player == %player.selectPlayer)
				say(%player, 0,
					"\tYou already have Admin Control.");
			else RocksControlGrabber(%player.selectPlayer);

			return;
		}


		if (%code == 5) {		// Add to BanList
			RocksBanListAdd(%player.selectPlayer);
			return;
		}


		if (%code == 9) {		// Ping selected player
			RocksPingPlayer(%player, %player.selectPlayer);
			return;
		}


		if (%code == 37) {		// Show selected player
			RocksShowPlayerInfo(%player, %player.selectPlayer, 1);
			return;
		}


		// Since it was none of the above, it might be one of the
		// vehicle codes below, so get the player's vehicle.
		//
		%player.selectVehicle =
		    playerManager::playerNumToVehicleId(%player.selectPlayer);


		if (!%player.selectVehicle) {
			say (%player, 0, "\t" @
				getName(%player.selectPlayer) @
				" has no Vehicle.");

			return;
		}
	}
	else %player.selectPlayer = 0;



	// Player.vehicle gets set either at the bottom of the previous
	// if-statement, which is the typical case, or because a clone
	// was selected in code 29... clones have no players.  But we
	// also have to make sure someone didn't destroy it.
	//
	if (%player.selectVehicle && isObject(%player.selectVehicle)) {

		if (%player.selectPlayer)
			%saystr = RocksInfoMaker(%player.selectPlayer);
		else %saystr = "\t" @ getHUDName(%player.selectVehicle);


		if (%code == 6) {		// SuperScan a vehicle
			if ($RocksSuperScan)
				RocksSuperScanner(%player,
					%player.selectVehicle);
			else say (%player, 0, "\tSuperScanner is off");
			return;
		}


		if (%code == 7) {		// Heal a vehicle
			RocksVehicleHeal(%player.selectVehicle);

			say (%player, 0, %saystr @ " Vehicle Healed");
			return;
		}


		if (%code == 8) {		// Destroy a vehicle
			RocksVehicleDestroy(%player.selectVehicle);

			say (%player, 0, %saystr @ " Vehicle Destroyed");
			%player.selectVehicle = 0;
			return;
		}


		if (%code == 37) {		// Show selected clone
			say (%player, 0, %saystr @ " (Selected)");
			return;
		}
	}
	else %player.selectVehicle = 0;


	say (%player, 0, "\tNo Player/Vehicle Selected or Unknown Admin Code");
}



// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// --------------------- Server Admin Extensions ----------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
function EXTPlayerOnAdd(%player)
{
	return(1);
}

function EXTPlayerOnRemove(%player)
{
}

function EXTTeamScore(%a)
{
	return(9999);
}

function EXTPlayerScore(%a)
{
	return(9999);
}

function EXTMissionStart()
{
}

function EXTVehicleOnDestroyed(%destroyed, %destroyer)
{
}

function EXTVehicleOnAdd(%vehicle)
{
}

function EXTVehicleOnScan(%scanned, %scanner)
{
}

function EXTStructureOnScan(%scanned, %scanner)
{
}

function EXTVehicleOnTargeted(%targeted, %targeter)
{
}

function EXTVehicleOnAttacked(%attackee, %attacker)
{
}

function EXTVehicleOnArrived(%vehicle, %where)
{
}

function EXTOnMissionPreLoad()
{
}

function EXTOnMissionEnd()
{
}

function EXTRules()
{
	return("");
}

function EXTcode(%admin, %player, %vehicle, %code,
		%code1, %code2, %code3, %code4, %code5,
		%code6, %code7, %code8, %code9, %code10)
{
	return(0);
}

function EXTServerRestart(%player, %code1, %code2, %code3, %code4,
		%code5, %code6, %code7, %code8, %code9, %code10)
{
	return(1);
}


// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// ------------------------ Game Initialization -----------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// These functions are called once for every game execution, via RocksDMlib.cs.
// Multiplayer games typically cycle through several games in a list.  These
// calls are made once for each of those games.
//
function RocksOnMissionPreLoad()
{

	if ($CmdLineServer == TRUE) {

		if ($RocksMapNumber == $RocksMapCount) {
			quit();
			return;
		}
		else $RocksMapNumber++;

		// Set the Team Colors if we're Dedicated, doing TDM,
		// and colors have been explicitly requested.  This lets
		// the Mission List change colors on the fly for each mission.
		//
		if (($server::TeamPlay == TRUE) &&
		    ($RocksTDMColors[$RocksMapNumber-1] != "")) {

			%colors = $RocksTDMColors[$RocksMapNumber-1];

			$server::AllowTeamPurple = floor(%colors / 1000);
			%colors %= 1000;

			$server::AllowTeamRed    = floor(%colors / 100);
			%colors %= 100;

			$server::AllowTeamBlue   = floor(%colors / 10);
			%colors %= 10;

			$server::AllowTeamYellow = %colors;
		}

		RocksLogEntry("New Game Start: " @ $missionName	@
			" (" @ $RocksMapNumber @ " of " @ $RocksMapCount @ ")");
	}
	else RocksLogEntry("New Game Start: " @ $missionName);


	$RocksAllowedTeams = 0;

	if ($server::AllowTeamPurple)
		$RocksAllowedTeams |= 8;

	if ($server::AllowTeamRed)
		$RocksAllowedTeams |= 4;

	if ($server::AllowTeamBlue)
		$RocksAllowedTeams |= 2;

	if ($server::AllowTeamYellow)
		$RocksAllowedTeams |= 1;


	if ($server::TeamPlay == TRUE) {

		RocksTDMColorSet(1);

		// remind players of the AutoTeam Mode
		//
		RocksTDMChoice(0);
	}


	$RocksGameCycle = TRUE;

	RocksRules();

	EXTOnMissionPreLoad();
}


// --------------------------------------------------------------------------
// Shut down the Monitor.  It starts again when the first player enters
// the field with a vehicle.  This is because the scheduler loses everything
// when the mission cycles.
//
function RocksOnMissionEnd()
{
	RocksMonitor(0);

	if ($RocksRandomThoughts)
		RRTStop();

	$RocksMissionFlag = 0;

	flushConsoleScheduler();

	if ($CmdLineServer == TRUE) {
		RocksLogEntry("Game Complete: " @ $missionName	@
			" (" @ $RocksMapNumber @ " of " @ $RocksMapCount @ ")");
	}
	else RocksLogEntry("Game Complete: " @ $missionName);


	if (($CmdLineServer == TRUE) && ($RocksMapNumber == $RocksMapCount))
		RocksLogEntry("Server Auto-Restart");
	else {
		// Reset all the Idle timers.  Covers the case for map cycles.
		//
		%count = playerManager::getPlayerCount();

		for (%i = 0; %i < %count; %i++) {
			%player = playerManager::getPlayerNum(%i);

			if (%player.admin)
				%player.idle = 0;
			else %player.idle = $RocksIdleTime;

			%player.teamname = "";
		}
	}


	EXTOnMissionEnd();
}


// --------------------------------------------------------------------------
// Initializes for a Skirmish Server and for Auto-error recovery.
//
function RocksAutoConfigInit()
{
	// will get forced to 1 if executing a Skirmish Autoconfig file
	//
	$RocksSkirmishGame = 0;

	// Only valid for Dedicated Servers
	//
	if ($CmdLineServer != TRUE)
		return;

	if (isFile($RocksAutoConfigFile))
		exec($RocksAutoConfigFile);


	// Restores the default for error recovery purposes
	//
	RocksWriteAutoConfig(0, 0, 0, 0, "", "", "", 0, 0, 0);


	if ($RocksSkirmishGame)
		exec($RocksServerFlipperFile);
}


//-----------------------------------------------------------------------*/
// This is a Map-List randomizer.  GroundPounder suggested this feature and
// after a little discussion with him, I agreed to code this up.  It's a nice
// little addition to the package.  Thanks GP.
//
// This only runs at Server-start or reset time, then it never runs again.
// Also, it only runs on Dedicated Servers.
//
// This function does two things with the map-list that is declared in the
// Server file.  It can select a map at random from the list and then rotate
// thru the list in the order that the list was setup in the Server File,
// or it can randomize the entire list.  Since these operations are logically
// mutually exclusive operations, it doesn't make sense to do both at the
// same time.  Accordingly, if both options are selected, the full-list
// randomization takes precedence.
// 
// Note that the RocksServerInit() function already counted the maps.
//
function RocksMapRandomizer()
{

	// Since RocksRandomMapList is a logical super-set of the other
	// option, we won't bother processing the RocksRandomMapStart
	// option if this one is enabled.
	//
	if ($RocksRandomMapList) {

		// first copy the list to a temporary storage
		//
		for (%i = 0; %i < $RocksMapCount; %i++) {
			%map[%i]    = $MissionCycling::Stage[%i];

			if ($RocksTDMColors[%i] != "")
				%colors[%i] = $RocksTDMColors[%i];
		}


		// now pull them out of the map-list at random and
		// plug them back into the $Mission list.
		//
		for (%i = 0; %i < $RocksMapCount; %i++) {

			// find one we haven't used yet
			//
			%rand = randomInt(0, ($RocksMapCount-1));
			while (%map[%rand] == "") {
				%rand = randomInt(0, ($RocksMapCount-1));
			}

			$MissionCycling::Stage[%i] = %map[%rand];

			if (%colors[%rand] != "")
				$RocksTDMColors[%i] = %colors[%rand];


			// clear it to mark it as used
			//
			%map[%rand]    = "";
			%colors[%rand] = "";

		}

		$server::Mission = $MissionCycling::Stage0;

		return;
	}


	// If RocksRandomMapStart is enabled, we will pick a map from the
	// list at random, but then we have to resort the list.  For example,
	// given ten maps #0 thru #9, if map #8 is chosen at random as the
	// first map, then we need to resort the map list so that the first
	// mission is #8, the second is #9, the third is #0, the forth is #1,
	// etc.  This is done by loading temporary variables with the new
	// list, then reloading the $Mission Map variables.
	//
	if (!$RocksRandomMapStart)
		return;

	%start = randomInt(0, ($RocksMapCount-1));


	for (%i = 0; %i < $RocksMapCount; %i++) {

		%map[%i]    = $MissionCycling::Stage[%start];

		if ($RocksTDMColors[%start] != "")
			%colors[%i] = $RocksTDMColors[%start];

		%start = (%start + 1) % $RocksMapCount;
	}


	// reload the Mission values with the sorted list
	//
	for (%i = 0; %i < $RocksMapCount; %i++) {

		$MissionCycling::Stage[%i] = %map[%i];

		if (%colors[%i] != "")
			$RocksTDMColors[%i] = %colors[%i];
	}

	$server::Mission = $MissionCycling::Stage0;
}



// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// ------------------------ Server Initialization ---------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// This function is called only one time when the server is first fired up.
// It is called from a file such as Server_DM_ALL.cs, or its equivalent.
//
function RocksServerInit()
{
	RocksAutoConfigInit();

	RocksAsciiInit();

	RocksPasswordInit();

	RocksLogInit();

	RocksBanListInit();

	RocksTDMInit(1);


	if (isFile($RocksInstallDir @ "\\" @ $RocksAdminExtensions))
		exec($RocksInstallDir @ "\\" @ $RocksAdminExtensions);


	// count the number of maps in the map-list
	//
	if ($CmdLineServer == TRUE) {
		%map = $MissionCycling::Stage0;

		for ($RocksMapCount = 0; %map != "";
		eval("%map = $MissionCycling::Stage" @ $RocksMapCount @ ";")) {
			$RocksMapCount++;
		}

		RocksMapRandomizer();

		if (($RocksResetMapLength > 0) &&
		    ($RocksResetMapLength < $RocksMapCount))
			$RocksMapCount = $RocksResetMapLength;

		$RocksMapNumber = 0;
	}


	if ($RAKeyEnable) {
		if (isFile("scripts\\RAKcfg.cs"))
			exec("scripts\\RAKcfg.cs");
		else $RAKeyEnable = FALSE;
	}

	if ($RocksSuperScan) {
		if (isFile($RocksInstallDir @ "\\RocksScanner.cs"))
			exec($RocksInstallDir @ "\\RocksScanner.cs");
		else $RocksSuperScan = FALSE;
	}

	if ($RocksRandomThoughts) {
		if (isFile($RocksInstallDir @ "\\RocksRandomThoughts.cs"))
			exec($RocksInstallDir @ "\\RocksRandomThoughts.cs");
		else $RocksRandomThoughts = FALSE;
	}


	RocksLogEntry("Server Start: " @ $server::Hostname);

	if ($RocksAdminColor == "BLUE")
		$RocksAdminColor = *IDSTR_TEAM_BLUE;
	else if ($RocksAdminColor == "PURPLE")
		$RocksAdminColor = *IDSTR_TEAM_PURPLE;
	else if ($RocksAdminColor == "YELLOW")
		$RocksAdminColor = *IDSTR_TEAM_YELLOW;
	else $RocksAdminColor = *IDSTR_TEAM_RED;

	$RocksServerCycle = 0;


	// Unbind the NONDedicated Package Load key to
	// prevent an accidental reload
	//
	unbind(keyboard0, make, alt, shift, "J");
}



// This makes the above function execute when this file is exec'd
//
RocksServerInit();

